/* #define	Xaw3d */

#include <stdio.h>

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Core.h>
#include <X11/Object.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>

#include "fileselect_select.h"

/* X11 stuff */

static	Display	*dpy ;
static	Atom	wm_protocols ;
static	Atom	wm_delete_window ;

/* Xt stuff */

static	XtAppContext	app_ctx ;

static	Widget	topLevel, form, dummyButton ;
static	Widget	fileSelect ;

static String fallback[] = {
  "FilePrompt.title:	Enter file name",
  "FilePrompt.baseTranslations: #override\n\
		<Message>WM_PROTOCOLS: WMProtocols()\n",
  NULL
};

static	void	OkCB(Widget cmd, XtPointer client, XtPointer call_data) ;

static	void	WMProtocols(Widget, XEvent *, String *, Cardinal *) ;

XtActionsRec	fileprompt_actions[] = {
  {"WMProtocols", WMProtocols},
} ;


static	char	usage[] =
"\nPrompt user for a filename, return result.\n\n usage: \nfileprompt\t[-test]\n\t[-fname 'human readable filter name, e.g. MS Word'] [-fmask 'pattern, e.g. *.doc'] [-dir initialdirpath] [-file initialvalue]\n [-title 'prompt string'] [-exist '0/1'] [-out 'output file for result string'] [-width <width> -height <height>]\n please give filtername 1st and corresponding filtermask 2nd throughout \n\n" ;

/* size constraints */
#define MAX_NUM_FILTERS 100
#define MAX_STR_LEN     1024

/* file name for result output */
static char    outname     [MAX_STR_LEN];
 
void write2outfile( char * result )
{
  FILE* outfile;

/*  printf("writing %s to %s\n", result, outname ); */
  if( strlen(outname) == 0 )
    return;

  outfile = fopen( outname, "w" );
  if( outfile == NULL )
    {
      printf( "couldn't open %s for writing\n", outname );
      return;
    }

  if( result == NULL )
    fprintf(outfile, "  ");
  else
    fprintf( outfile, result );

  fprintf( outfile, "\n" );

  fclose( outfile );
}

int
prompt_for_file( char * outname,
		 char * caption, char * filename, char * initial_path,
                 char ** filternames, char ** filtermasks,
                 int numFilters,
                 int shouldExist,
		 int widgetWidth, /* -1 -> don't resize */
	         int widgetHeight )
{
  int     i;
  int     argc = 0;
  char ** argv = NULL;

  topLevel = XtVaAppInitialize(&app_ctx, "FilePrompt", NULL,0,
			       &argc, argv, fallback, 
			       XtNtitle, (caption == NULL) ? "Select Filename" : caption, 
			       0) ;
  dpy = XtDisplay(topLevel) ;

  
  /* set up to handle quits */
  
  XtAppAddActions(app_ctx,
		  fileprompt_actions, XtNumber(fileprompt_actions));
  XtOverrideTranslations (topLevel, 
			  XtParseTranslationTable ("<Message>WM_PROTOCOLS: WMProtocols()"));  
  
  fileSelect = XtVaCreateManagedWidget("Sampleselector",
				       fileSelectWidgetClass, topLevel, /* ARGV, ARGC */0) ;
  if( widgetWidth != -1 )
	{
		XtResizeWidget( fileSelect, widgetWidth, widgetHeight, 3 );
	}
  XtAddCallback(fileSelect, XtNcallback, OkCB, 0) ;
  FileSelectSetInitial(fileSelect,
                       filename, initial_path,
                       filternames, filtermasks, numFilters,
                       shouldExist);    
  XtRealizeWidget(topLevel) ;
  
  /* tell window system we're willing to handle window-delete messages */
  wm_protocols = XInternAtom(dpy, "WM_PROTOCOLS", False);
  wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
  (void) XSetWMProtocols(dpy, XtWindow(topLevel), &wm_delete_window,1);
  
  XtAppMainLoop(app_ctx) ;
  
/*  printf("byebye (2) ...\n"); */
  write2outfile( NULL );
  exit(0) ;
}


main(int argc, char **argv)
{
  int i, n, t;
  char    caption     [MAX_STR_LEN];
  char    initialname [MAX_STR_LEN];
  char    initialpath [MAX_STR_LEN];
  char    filternames [MAX_NUM_FILTERS] [MAX_STR_LEN];
  char    filtermasks [MAX_NUM_FILTERS] [MAX_STR_LEN];
  char*   processed_filternames [MAX_NUM_FILTERS];
  char*   processed_filtermasks [MAX_NUM_FILTERS];
  int     numfilters;
  int     shouldexist;
  int     widgetWidth=-1;
  int     widgetHeight=-1;

  /* initialize */
  caption[0]     = '\0';
  initialname[0] = '\0';
  initialpath[0] = '\0';
  numfilters     = 0;
  shouldexist    = 0;
  outname[0]     = '\0';

  /* parse the command line options */
  
  while( --argc > 0 ) {
    ++argv ;
    if( strcmp(*argv, "-fname") == 0  &&  --argc > 0 )
      {
/*	printf("filter name: %s\n", *++argv ); */
	++argv;
	strcpy( filternames[numfilters], *argv );
      }
    else if( strcmp(*argv, "-fmask") == 0 && --argc > 0 )
      {
/*	printf("filter mask: %s\n", *++argv ); */
	++argv;
	strcpy( filtermasks[numfilters], *argv );
	numfilters ++;
      }
    else if( strcmp(*argv, "-dir") == 0 && --argc > 0 )
      {
/*	printf("initial path: %s\n", *++argv ); */
	++argv;
	strcpy(initialpath, *argv);
      }
    else if( strcmp(*argv, "-file") == 0 && --argc > 0 )
      {
/*	printf("initial file name: %s\n", *++argv ); */
	++argv;
	strcpy(initialname, *argv);
      }
    else if( strcmp(*argv, "-title") == 0 && --argc > 0 )
      {
/*	printf("title: %s\n", *++argv ); */
	++argv;
	strcpy(caption, *argv);
      }
    else if( strcmp(*argv, "-exist") == 0 && --argc > 0 )
      {
/* 	printf("file should exist: %s\n", *++argv ); */
	++argv;
	if( strcmp(*argv, "0" ) == 0 )
	  shouldexist=0;
	else
	  shouldexist=1;
/*	printf("again: %d\n", shouldexist ); */
      }
    else if( strcmp(*argv, "-out") == 0 && --argc > 0 )
      {
/* 	printf("file name for output: %s\n", *++argv ); */
	++argv;
	strcpy(outname, *argv);

	/* ensure to be prepared for cancellation */
	write2outfile( NULL );
      }
    else if( strcmp(*argv, "-width") == 0 && --argc > 0 )
      {
	++argv;
	sscanf(*argv, "%d", &widgetWidth);
      }
    else if( strcmp(*argv, "-height") == 0 && --argc > 0 )
      {
	++argv;
	sscanf(*argv, "%d", &widgetHeight);
      }
    else if( strcmp(*argv, "-test") == 0 )
      {
	exit( 0 );
      }
    else {
      fprintf(stderr, usage) ;
      exit(2) ;
    }
  }

  /* process filters to convert to convinient format */
 
  n = 0;
  t = 0;
  for( i=0; i<numfilters; i++ )
    {
/*      printf( "name for %d. file filter: %s\n", i, filternames[i] ); */
      processed_filternames[i] = (char *) &filternames[i];
/*      printf( "mask for %d. file filter: %s\n", i, filtermasks[i] ); */
      processed_filtermasks[i] = (char *) &filtermasks[i];
    }   

  if( widgetHeight == -1 )
	widgetWidth = -1;

  prompt_for_file( outname, caption, initialname, initialpath, processed_filternames, processed_filtermasks, 
		   numfilters, shouldexist, widgetWidth, widgetHeight );
}


static	void
OkCB(Widget cmd, XtPointer client, XtPointer call_data)
{
  if( call_data == NULL )
    {
      write2outfile( NULL );
      exit(1) ;
    }
  
/*  printf("%s\n", (String)call_data) ; */
  write2outfile( (String)call_data );
  exit(0) ;
}


static	void
WMProtocols(w, ev, params, nparams)
	Widget	w ;
	XEvent	*ev ;
	String	*params ;
	Cardinal *nparams ;
{
	if( ev->type == ClientMessage  &&
	    ev->xclient.message_type == wm_protocols  &&
	    ev->xclient.data.l[0] == wm_delete_window )
	{
	  if( w == topLevel )
	    {
/*	      printf("byebye (3)...\n"); */
	      exit(1) ;
	    }
	}
}





















